import time
from collections.abc import Sequence

import grpc

from bencherscaffold.dual_stack_service import grpc_target
from bencherscaffold.protoclasses.bencher_pb2 import (
    BenchmarkRequest, EvaluationResult, Point, Benchmark,
    BenchmarkType, Value, ValueType,
)
from bencherscaffold.protoclasses.bencher_pb2_grpc import BencherStub


class BencherClient:
    def __init__(
            self,
            address: str = '127.0.0.1',
            port: int = 50051,
            max_retries: int = 10,
            wait_time: int = 5,
    ):
        """
        Initializes the BencherClient with the given parameters.
        Args:
            address: The address of the server. Can be an IP address/hostname
                or a Unix socket path (e.g. "unix:///tmp/bencher.sock").
            port: The port number of the server. Ignored for Unix sockets.
            max_retries: The maximum number of retries for the request.
            wait_time: The time to wait between retries in seconds.
        """
        self.channel = grpc.insecure_channel(grpc_target(address, port))
        self.stub = BencherStub(self.channel)
        self.max_retries = max_retries
        self.wait_time = wait_time

    def evaluate_point(
            self,
            benchmark_name: str,
            point: Sequence[Value]
    ) -> float:
        """
        Evaluates a point in the benchmark space.
        This method sends a request to the server to evaluate a specific point in the benchmark space.
        It retries the request if it fails, up to the maximum number of retries specified.
        If the request fails after the maximum number of retries, it raises the last exception encountered.
        The point must be a sequence of floats, and its length must match the number of dimensions of the benchmark.

        Args:
            benchmark_name: The name of the benchmark to evaluate.
            point:  A sequence of floats representing the point in the benchmark space to evaluate.

        Returns:
            The evaluated value of the point in the benchmark space.

        """

        if all(p.type == ValueType.CONTINUOUS for p in point):
            benchmark_type = BenchmarkType.PURELY_CONTINUOUS
        elif all(p.type == ValueType.BINARY for p in point):
            benchmark_type = BenchmarkType.PURELY_BINARY
        elif all(p.type == ValueType.INTEGER for p in point):
            benchmark_type = BenchmarkType.PURELY_ORDINAL_INT
        elif all(p.type == ValueType.CATEGORICAL for p in point):
            benchmark_type = BenchmarkType.PURELY_CATEGORICAL
        else:
            benchmark_type = BenchmarkType.MIXED

        benchmark = Benchmark(
            name=benchmark_name,
            type=benchmark_type,
        )

        request = BenchmarkRequest(
            benchmark=benchmark,
            point=Point(
                values=point,
            ),

        )
        for n_retry in range(self.max_retries):
            try:
                response: EvaluationResult = self.stub.evaluate_point(request)
                return response.value
            except grpc.RpcError as e:
                if n_retry < self.max_retries - 1:
                    print(f"Retrying... {n_retry + 1}/{self.max_retries}")
                    time.sleep(self.wait_time)
                else:
                    raise e



from concurrent.futures import ThreadPoolExecutor

import grpc
import logging
import os

from bencherscaffold.protoclasses import second_level_services_pb2_grpc
from bencherscaffold.protoclasses.second_level_services_pb2_grpc import SecondLevelBencherServicer


class GRCPService(SecondLevelBencherServicer):

    def __init__(
            self,
            host: str = "127.0.0.1",
            port: int = 50000,
            n_cores: int = os.cpu_count()
    ):
        """
        Args:
            host: The host name to use for the connection. If not provided, the default hostname is set to '127.0.0.1' (ipv4 localhost).
            port: The port number to use for the connection. If not provided, the default port is set to 50000.
            n_cores: The number of CPU cores to use for parallel processing. If not provided, it defaults to os.cpu_count().

        """
        self.host = host
        self.port = port
        self.n_cores = n_cores

    def serve(
            self
    ):
        """
        Serves the gRPC server and starts listening on the specified port.

        Args:
            self (GRCPService): The instance of the SecondLevelBencherServicer class.

        Returns:
            None

        Raises:
            None
        """
        logging.basicConfig()
        n_cores = self.n_cores
        port = str(self.port)
        server = grpc.server(ThreadPoolExecutor(max_workers=n_cores))
        second_level_services_pb2_grpc.add_SecondLevelBencherServicer_to_server(self, server)
        server.add_insecure_port(f"{self.host}:{port}")
        server.start()
        print("Server started, listening on " + port)
        server.wait_for_termination()



from argparse import ArgumentParser
from concurrent.futures import ThreadPoolExecutor
import os
from typing import Iterable, Tuple, Optional, List, Sequence

import grpc
import logging

from bencherscaffold.protoclasses import second_level_services_pb2_grpc
from bencherscaffold.protoclasses.grcp_service import GRCPService as _BaseGRCPService

_DEFAULT_HOSTS: Tuple[str, str] = ("0.0.0.0", "[::]")
_DEFAULT_LISTEN_EXAMPLES: Tuple[str, str, str] = ("127.0.0.1", "[::]", "unix:/tmp/benchmark.sock")


def _normalize_hosts(listen_hosts: Optional[Iterable[str]]) -> Tuple[str, ...]:
    """
    Normalizes a list of listener hosts, handling None and removing duplicates/empty strings.
    This function has been updated to use Optional[Iterable[str]] instead of Iterable[str] | None
    and List/Tuple for consistency with Python 3.8 typing.
    """
    if listen_hosts is None:
        listen_hosts = _DEFAULT_HOSTS
    # Explicitly use List from typing for Python 3.8 compatibility
    normalized: List[str] = []
    for host in listen_hosts:
        candidate = host.strip()
        if candidate and candidate not in normalized:
            normalized.append(candidate)
    return tuple(normalized)


def grpc_target(host_or_address: str, port: int) -> str:
    if host_or_address.startswith("unix:"):
        return host_or_address
    return f"{host_or_address}:{port}"


def parse_listen_entries(raw_value: Optional[str]) -> Optional[Tuple[str, ...]]:
    if raw_value is None:
        return None
    return _normalize_hosts(raw_value.split(","))


def resolve_listen_entries(
        cli_entries: Optional[Iterable[str]],
        env_var: Optional[str] = None,
        default: Optional[Iterable[str]] = None,
) -> Tuple[str, ...]:
    if cli_entries is not None:
        return _normalize_hosts(cli_entries)

    if env_var is not None:
        env_entries = parse_listen_entries(os.environ.get(env_var))
        if env_entries is not None:
            return env_entries

    return _normalize_hosts(default)


def add_listen_argument(
        parser: ArgumentParser,
        env_var: str,
        option: str = "--listen-host",
        dest: str = "listen_hosts",
        value_name: str = "Host or gRPC address",
        examples: Sequence[str] = _DEFAULT_LISTEN_EXAMPLES,
) -> None:
    example_str = ", ".join(examples)
    parser.add_argument(
        option,
        dest=dest,
        action="append",
        default=None,
        help=f"{value_name} to listen on (for example {example_str}). "
             f"Repeat to bind multiple values. Can also be set via the {env_var} "
             "environment variable as a comma-separated list.",
    )


class DualStackGRCPService(_BaseGRCPService):

    def __init__(
            self,
            *,
            listen_hosts: Optional[Iterable[str]] = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._listen_hosts = _normalize_hosts(listen_hosts)

    @property
    def listen_hosts(self) -> Tuple[str, ...]:
        return self._listen_hosts

    def serve(self):
        logging.basicConfig()
        server = grpc.server(ThreadPoolExecutor(max_workers=self.n_cores))
        second_level_services_pb2_grpc.add_SecondLevelBencherServicer_to_server(self, server)
        bound = 0
        for host in self._listen_hosts:
            bound += server.add_insecure_port(grpc_target(host, self.port))
        if bound == 0:
            raise RuntimeError(f"Could not bind benchmark service on port {self.port} for hosts {self._listen_hosts}")
        addresses = ", ".join(self._listen_hosts)
        print(f"Server started, listening on {self.port} via {addresses}")
        server.start()
        server.wait_for_termination()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bencherscaffold/protos/second_level_services.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from bencherscaffold.protoclasses import bencher_pb2 as bencherscaffold_dot_protos_dot_bencher__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2bencherscaffold/protos/second_level_services.proto\x1a$bencherscaffold/protos/bencher.proto2N\n\x12SecondLevelBencher\x12\x38\n\x0e\x65valuate_point\x12\x11.BenchmarkRequest\x1a\x11.EvaluationResult\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bencherscaffold.protos.second_level_services_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_SECONDLEVELBENCHER']._serialized_start=92
  _globals['_SECONDLEVELBENCHER']._serialized_end=170
# @@protoc_insertion_point(module_scope)



from bencherscaffold.protoclasses import bencher_pb2 as _bencher_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class BenchmarkType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PURELY_CONTINUOUS: _ClassVar[BenchmarkType]
    PURELY_BINARY: _ClassVar[BenchmarkType]
    PURELY_CATEGORICAL: _ClassVar[BenchmarkType]
    PURELY_ORDINAL_REAL: _ClassVar[BenchmarkType]
    PURELY_ORDINAL_INT: _ClassVar[BenchmarkType]
    MIXED: _ClassVar[BenchmarkType]

class ValueType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CONTINUOUS: _ClassVar[ValueType]
    BINARY: _ClassVar[ValueType]
    INTEGER: _ClassVar[ValueType]
    CATEGORICAL: _ClassVar[ValueType]
PURELY_CONTINUOUS: BenchmarkType
PURELY_BINARY: BenchmarkType
PURELY_CATEGORICAL: BenchmarkType
PURELY_ORDINAL_REAL: BenchmarkType
PURELY_ORDINAL_INT: BenchmarkType
MIXED: BenchmarkType
CONTINUOUS: ValueType
BINARY: ValueType
INTEGER: ValueType
CATEGORICAL: ValueType

class Value(_message.Message):
    __slots__ = ("type", "value")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    type: ValueType
    value: float
    def __init__(self, type: _Optional[_Union[ValueType, str]] = ..., value: _Optional[float] = ...) -> None: ...

class Benchmark(_message.Message):
    __slots__ = ("name", "type", "description")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: BenchmarkType
    description: str
    def __init__(self, name: _Optional[str] = ..., type: _Optional[_Union[BenchmarkType, str]] = ..., description: _Optional[str] = ...) -> None: ...

class BenchmarkRequest(_message.Message):
    __slots__ = ("benchmark", "point")
    BENCHMARK_FIELD_NUMBER: _ClassVar[int]
    POINT_FIELD_NUMBER: _ClassVar[int]
    benchmark: Benchmark
    point: Point
    def __init__(self, benchmark: _Optional[_Union[Benchmark, _Mapping]] = ..., point: _Optional[_Union[Point, _Mapping]] = ...) -> None: ...

class Point(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedCompositeFieldContainer[Value]
    def __init__(self, values: _Optional[_Iterable[_Union[Value, _Mapping]]] = ...) -> None: ...

class EvaluationResult(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: float
    def __init__(self, value: _Optional[float] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bencherscaffold/protos/bencher.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$bencherscaffold/protos/bencher.proto\"0\n\x05Value\x12\x18\n\x04type\x18\x01 \x01(\x0e\x32\n.ValueType\x12\r\n\x05value\x18\x02 \x01(\x01\"L\n\tBenchmark\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1c\n\x04type\x18\x02 \x01(\x0e\x32\x0e.BenchmarkType\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\"H\n\x10\x42\x65nchmarkRequest\x12\x1d\n\tbenchmark\x18\x01 \x01(\x0b\x32\n.Benchmark\x12\x15\n\x05point\x18\x02 \x01(\x0b\x32\x06.Point\"\x1f\n\x05Point\x12\x16\n\x06values\x18\x01 \x03(\x0b\x32\x06.Value\"!\n\x10\x45valuationResult\x12\r\n\x05value\x18\x01 \x01(\x01*\x8d\x01\n\rBenchmarkType\x12\x15\n\x11PURELY_CONTINUOUS\x10\x00\x12\x11\n\rPURELY_BINARY\x10\x01\x12\x16\n\x12PURELY_CATEGORICAL\x10\x02\x12\x17\n\x13PURELY_ORDINAL_REAL\x10\x03\x12\x16\n\x12PURELY_ORDINAL_INT\x10\x04\x12\t\n\x05MIXED\x10\x05*E\n\tValueType\x12\x0e\n\nCONTINUOUS\x10\x00\x12\n\n\x06\x42INARY\x10\x01\x12\x0b\n\x07INTEGER\x10\x02\x12\x0f\n\x0b\x43\x41TEGORICAL\x10\x03\x32\x43\n\x07\x42\x65ncher\x12\x38\n\x0e\x65valuate_point\x12\x11.BenchmarkRequest\x1a\x11.EvaluationResult\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bencherscaffold.protos.bencher_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_BENCHMARKTYPE']._serialized_start=311
  _globals['_BENCHMARKTYPE']._serialized_end=452
  _globals['_VALUETYPE']._serialized_start=454
  _globals['_VALUETYPE']._serialized_end=523
  _globals['_VALUE']._serialized_start=40
  _globals['_VALUE']._serialized_end=88
  _globals['_BENCHMARK']._serialized_start=90
  _globals['_BENCHMARK']._serialized_end=166
  _globals['_BENCHMARKREQUEST']._serialized_start=168
  _globals['_BENCHMARKREQUEST']._serialized_end=240
  _globals['_POINT']._serialized_start=242
  _globals['_POINT']._serialized_end=273
  _globals['_EVALUATIONRESULT']._serialized_start=275
  _globals['_EVALUATIONRESULT']._serialized_end=308
  _globals['_BENCHER']._serialized_start=525
  _globals['_BENCHER']._serialized_end=592
# @@protoc_insertion_point(module_scope)
